#include <stdio.h>
#include <fcntl.h>
#include <errno.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <ggi/graphdev.h>
#include <ggi/graphlib.h>

void 
init_textmode(ggi_mode* mode, int x, int y, int fontx, int fonty)
{
  mode->visible.x = mode->virtual.x = x * fontx; 
  mode->visible.y = mode->virtual.y = y * fonty;
  mode->size.x = mode->size.y = 0;
  mode->textgrid.x = fontx; mode->textgrid.y = fonty;
  mode->graphtype = GT_TEXT;
}

void 
printf_mode(ggi_mode *mode)
{
  printf("mode = \n{\n\t{ %i, %i },\t/* visible\t*/\n", mode->visible.x, mode->visible.y);
  printf("\t{ %i, %i },\t/* virtual\t*/\n",mode->virtual.x, mode->virtual.y);
  printf("\t{ %i, %i },\t/* size[mm]\t*/\n",mode->size.x, mode->size.y);
  printf("\t%i,\t\t/* graphtype\t*/\n", mode->graphtype);
  printf("\t{ %i, %i }\t/* textgrid\t*/\n};\n",mode->textgrid.x, mode->textgrid.y);
}

int main(int argc,char **argv)
{
  int x,y,cols,col;
  ggi_mode mode = {  { 320, 200 }, 
                     { 320, 200 },
		     {   0,   0 },
		     GT_8BIT,
		     {   9,  16 }
		 };
 
  if (argc!=2 || sscanf(argv[1],"%dx%dx%d",&x,&y,&cols)!=3) {
    fprintf(stderr,"Usage:\n\n%s <width>x<height>x<depth>\n",argv[0]);
    return 1;
  }
  mode.visible.x=mode.virtual.x=x;
  mode.visible.y=mode.virtual.y=y;
  switch (cols) {
    case   1: cols=GT_1BIT; break;
    case   4: cols=GT_4BIT; break;
    case   8: cols=GT_8BIT; break;
    case  15: cols=GT_15BIT; break; 
    case  16: cols=GT_16BIT; break;
    case  24: cols=GT_24BIT; break; 
    case  32: cols=GT_24BIT; break; 
    default: fprintf(stderr,"%s: %d is an unsupported colordepth\n",
                     argv[0],cols);
             return 1;
  }
  mode.graphtype=cols;

  ggi_init ();

  if (ggi_setmode (&mode))
    {
      printf ("mode not ok.\n");
      ggi_exit();
      exit(1);
    }

  while (ggi_getch()!=27)
   {
     int x,y,i,maxx,maxy;

     for (y = 0; y < mode.virtual.y; y++)
       for (x = 0; x < mode.virtual.x; x++)
	 {
	   ggi_setcolor(x+y*257);
	   ggi_drawpixel(x, y);
	 }
     ggi_getch();
     ggi_setcolor(-1);
     for (x = 0; x < mode.virtual.x; x++)
       for (y = 0; y < mode.virtual.y; y++)
	 ggi_drawpixel(x, y);

   } 
  ggi_exit();
  return 0;
}
