/*\
|*| --------------------------------------------------------------------------
|*|	S3 Vision864 chipset driver
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995	Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
|*|
|*|	This code was derived from the following source of information:
|*|	[1]	S3 Vision864 Data Book, (c) 1994 S3 Incorporated
\*/

/* !!!	[1] states that DCLK must be greater than max(10 MHz, 0.5*SystemCLK)
 * !!!	(which might be up to 25 MHz on VL bus) to ensure valid writes to the
 * !!!	clock chip. To allow for 320 pixel horizontal (ca. 400 pixel htotal) 
 * !!!	resolution on a standard VGA monitor (25.175 kHz) we need DCLK down to
 * !!!	10 MHz. My miroCrystal20SV didn't seem to have any problems so 
 * !!!	chipinfo.dclk.min is set to 10 MHz. Hopefully all is well... 
 */

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#define	DRIVER_VERSION	"0.00"

#undef	UNDERLINE	/* define to enable underline (attr 01h 09h 81h 89h) */
/* #undef	DEBUG */
#define	DEBUG_IO

/*\
|*|	Kernel includes
\*/

#include <linux/mm.h>
#include <linux/mman.h>
#include <asm/pgtable.h>
  
extern void request_region(unsigned int from, unsigned int num,
                           const char *name);
extern int  check_region  (unsigned int from, unsigned int num);
extern void release_region(unsigned int from, unsigned int num);

/*\
|*|     GGI includes
\*/

#include <ggi/graphdev.h>

#define	VISION864
#include "S3/Vision.h"
#include "s3_io.inc"

static struct kgi_chipset
chipinfo =
{
	"S3 Incorporated",	/* manufacturer		*/
	"Vision864",		/* model (detected)	*/
	GM_TEXT | GM_4BIT  |	/* supported modes	*/
	GM_8BIT | GM_15BIT |
	GM_16BIT| GM_32BIT,
	0,
	{ 2048, 2048 },		/* maximum resolution	*/

	0,			/* RAM size (detected)	*/
	{ 10000000, 95000000 }	/* DCLK limits		*/
};


static struct statetable	/* this table holds the default bits */
default_state =			/* for text & graphics mode, will be */
{				/* filled in during mode_init        */
  0x23, 0x00,					/* MISC, FCTRL */
  {		
     0x03, 0x00, 0x0F, 0x20, 0x02			/* SEQ */
  },
  {
     0x00, 0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x10,	/* CRT */
     0x00, 0x40, 0x00, 0x1F, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x1F, 0x00, 0x00, 0xE3,		/* 0x1x */
     0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x2x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x01, 0x00, 0x00, 0x10, 0x00, 0x00, 0x00,		/* 0x3x */
     0x48, 0xA5, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x4x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x38, 0x00, 0x00, 0x00,		/* 0x5x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x40, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x6x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,

     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,		/* 0x7x */
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
  },
  {
     0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,	/* GRC */
     0xFF
  },
  {
     0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* ATC */
     0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F,
     0x84, 0x00, 0x0F, 0x00, 0x00
  }
};

static unsigned long text_buffer = 0;
static unsigned long graph_buffer = 0;

static int
detect_Vision86x(void)
{
  #define COPYDEFAULT(X) (default_state.CRT[X] = inbCRT(X))
  register unsigned char foo;
  unsigned long len;

  foo = COPYDEFAULT(0x30);
  switch(foo & CR30_CHIPID)
   {
     case CR30_Vision864ID:		/* a 864 as stated in chipinfo */
	foo &= CR30_REVISION;
	break;
#if 0
     case CR30_Vision968ID:
	if (inbCRT(CRT_CHIPID_HIGH) == 0x88)
          if (inbCRT(CRT_CHIPID_LOW) == 0xF0)
	    {
	      chipinfo.model[8] = '8';
/* !!!	      chipinfo.modes |= GM_24BIT;	not supported yet. */
	      foo = inbCRT(CRT_CHIP_REV);
	      break;
	    }
#endif

     default: 
 	printk("  No Vision864 detected (ID = %.2x).\n", foo & CR30_CHIPID);
 	return -ECHIPSET;
   }
  TRACE(printk("  %s rev %i detected.\n", chipinfo.model, foo));

  foo = COPYDEFAULT(0x36);
   switch(foo & CR36_MEMSIZE_MASK)
    {
      case CR36_1MB: chipinfo.ramsize = 1;  len = 1; 	break;
      case CR36_2MB: chipinfo.ramsize = 2;  len = 2;	break;
      case CR36_3MB: chipinfo.ramsize = 3;  len = 8;	break;
      case CR36_4MB: chipinfo.ramsize = 4;  len = 8;	break;
      case CR36_6MB: chipinfo.ramsize = 6;  len = 8;	break;
      case CR36_8MB: chipinfo.ramsize = 8;  len = 8;	break;
      default:  chipinfo.ramsize = 1; len = 1;
                printk("%s: can't determine DRAM size, assume 1MB.\n",__FILE__);
    }
   #ifdef DEBUG
    switch(foo & CR36_RAMMASK)
     {
       case CR36_FASTPAGE: printk("  %i MB FPM RAM, ", chipinfo.ramsize); break;
       case CR36_EDO:      printk("  %i MB EDO RAM, ", chipinfo.ramsize); break;
       default:		   printk("  %i MB unknown RAM, ", chipinfo.ramsize);
     }
    switch(foo & CR36_BUSMASK)
     {
       case CR36_PCI      :   printk("PCI bus\n");        break;
       case CR36_VESALOCAL:   printk("Vesa Local bus\n"); break;
       default:               printk("unknown bus\n");
     }
   #endif
   if (chipinfo.ramsize ==  1)		/* limit DCLK for 1MB configurations */
     chipinfo.dclk.max = 75000000;
   chipinfo.ramsize *= 0x100000; /* convert MB to Byte */

  /*\
  !!!	This is to get physical memory. I do not feel very well the way this
  !!!	is done right now. Especially for PCI cards more care must be taken. 
  \*/

  fb_text_len = 0x8000;		/* we map 32K at 0x000b8000 */
  fb_text = kgi_get_physical_mem(&text_buffer, 0x000B8000, 0x000B8000, 
                                 fb_text_len, fb_text_len, chipinfo.model);
  if (! fb_text)
    {
      printk("%s: no physical address space (text framebuffer).\n", __FILE__);
      return -ECHIPSET;
    }

  fb_graph_len = len*0x100000; /* 1,2 or 8 MB anywhere in physical memory */
  fb_graphics = kgi_get_physical_mem(&graph_buffer, 0x00000000, 0xFFFFFFFF,
                                    fb_graph_len, fb_graph_len, chipinfo.model);
#if 0
  fb_graph_len = 0x10000;
  fb_graphics = kgi_get_physical_mem(&graph_buffer, 0x000A0000, 0x000A0000,
                                    fb_graph_len, fb_graph_len, chipinfo.model);
#endif
  if (! fb_graphics)
    {
      printk("%s: no physical address space (graphics framebuffer).\n",__FILE__);
/* !!!  kgi_free_physical_mem(fb_text, fb_text_len); */
      return -ECHIPSET;
    }

  printk("text: 0x%p graph: 0x%p\n", fb_text, fb_graphics);
  printk("buf:  0x%lx graph: 0x%lx\n", text_buffer, graph_buffer);

  COPYDEFAULT(CRT_CONFIG2);
  default_state.CRT[0x3A] |= (inbCRT(0x3A) & (CR3A_PCIMASK | CR3A_REFRESHMASK));
  default_state.CRT[0x40] |= (inbCRT(0x40) & (CR40_BUSMASK));
  default_state.CRT[0x58] |= (inbCRT(0x58) & (CR58_VL_T1_LATCH));
/* ??? default_state.CRT[0x5C] |= (inbCRT(0x5C) & (CR5C_GOUT_MASK)); */
  COPYDEFAULT(CRT_MEMCONTROL3);
  COPYDEFAULT(CRT_MEMCONTROL4);
  COPYDEFAULT(CRT_MEMCONTROL5);
  default_state.CRT[0x66] |= (inbCRT(0x66) & (CR66_PCI_DISCONNECT));
  COPYDEFAULT(CRT_CONFIG3);

  default_state.CRT[0x59] = (graph_buffer >> 24) & 0xFF;
  default_state.CRT[0x5A] = (graph_buffer >> 16) & 0xFF;
  switch (len)
   {
     case 1:	default_state.CRT[0x58] |= CR58_LAW_1MB; break;
     case 2:	default_state.CRT[0x58] |= CR58_LAW_2MB; break;
     case 8:	default_state.CRT[0x58] |= CR58_LAW_8MB; break;
   }

  return EOK;
  #undef COPYDEFAULT
}

#define	SETFONT 0
#define	GETFONT 1

static void 
getsetfont(int get, int no, ggi_font *font)
{
  IS_CRITICAL;
  unsigned char SR2, SR4, GR4, GR5, GR6;
  int i;
  char *fbuf,*data;

  if (font == NULL) return;

  fbuf = (char*) (0x000A0000 + (no * 0x2000) + (font->first << 5));
  data = font->data;

  ENTER_CRITICAL;
    SR2 = inbSEQ(0x02); SR4 = inbSEQ(0x04);	/* save register state        */
    GR4 = inbGRC(0x04); GR5 = inbGRC(0x05); GR6 = inbGRC(0x06);

    outbCRT(inbCRT(0x3A) | CR3A_HISPEEDTXT, 0x3A);
    outbSEQ(SR02_PLANE2, 0x02);			/* font data goes to plane 2  */
    outbSEQ(SR04_256K_ACCESS | SR04_NO_ODDEVEN, 0x04);

    outbGRC(2, 0x04);				/* read plane 2		      */
    outbGRC(GR5_WRITEMODE0, 0x05);
    outbGRC(GR6_MAP_A0_64 | GR6_GRAPHMODE, 0x06);
  LEAVE_CRITICAL;
  /* NOTE: no printk()'s are allowed in here unless klogd is active ! */

  if (get)
    {						/* read out font data         */
      if (data)
        for(i = font->first; i <= font->last; i++)
         {
           memmove(data, fbuf, font->ysize);
           fbuf += 32; data += font->ysize;
         }
    }
   else
    if (data)					/* font data valid ?          */
      {
        register int cnt = 32 - font->ysize;

        for(i = font->first; i <= font->last; i++)
         {
           memmove(fbuf, data, font->ysize);
           memset (fbuf + font->ysize, 0, cnt);
           fbuf += 32; data += font->ysize; 
         };
      }
     else					/* clear font data            */
      memset(fbuf, 0, (font->last - font->first + 1) << 5);

  ENTER_CRITICAL;
    outbCRT(  (inbCRT(0x0A)   & ~CR0A_CURSOR_START_MASK)/* setup cursor       */
            | (font->cr_start &  CR0A_CURSOR_START_MASK), 0x0A);
    outbCRT(  (inbCRT(0x0B)   & ~CR0B_CURSOR_END_MASK)
            | (font->cr_end   &  CR0B_CURSOR_END_MASK), 0x0B);
    #ifndef UNDERLINE
      font->underline = CR14_UNDERLINE_MASK;
    #endif
    outbCRT(  (inbCRT(0x14)   & ~CR14_UNDERLINE_MASK)	/* setup underline    */
            | (font->underline&  CR14_UNDERLINE_MASK), 0x14);

    outbSEQ(SR2, 0x02); outbSEQ(SR4, 0x04);		/* restore registers  */
    outbGRC(GR4, 0x04); outbGRC(GR5, 0x05); outbGRC(GR6, 0x06);
    outbCRT(inbCRT(0x3A) & ~CR3A_HISPEEDTXT, 0x3A); 
  LEAVE_CRITICAL;
}

/*\
|*|	GGI interface
\*/

int
kgi_chipset_init(void)
{
  TRACE(printk(" S3 864 driver "DRIVER_VERSION"\n"));

  outb(S3_DISABLE, S3_SUBSYSTEM);	/* chip wakeup		*/
  outb(S3_WAKEUP,  S3_SETUP);
  outb(S3_ENABLE,  S3_SUBSYSTEM);

  outbMISC(inb(MISCr));			/* adjust IO base	*/

  s3_unlock();

  if(! detect_Vision86x())
    {
      register unsigned char foo;
      
      /* select CR42 to drive the clock select lines. */
      foo = inb(MISCr);
      if((foo & MISC_CLOCK_MASK) != MISC_EXT_CLOCK)
        { 
          outbCRT(((foo & MISC_CLOCK_MASK) >> 2), 0x42);
          outbMISC((foo & ~MISC_CLOCK_MASK) | MISC_EXT_CLOCK);
        }

      if (! check_region(S3IO_BASE, S3IO_REGION))
        request_region(S3IO_BASE, S3IO_REGION, "Vision964");
       else   
        printk("%s: warning: failed to allocate IO region.\n",__FILE__);

      return EOK; 
    }
   else
    return -ECHIPSET;
}

void
kgi_chipset_done(void)
{
  release_region(S3IO_BASE, S3IO_REGION);
  TRACE(printk("S3 Chipset Driver removed\n"));
  s3_lock();
}

struct kgi_chipset 
*kgi_chipset_get_info (void)
{
  return &chipinfo;
}

static void 
adjust_timing(struct kgi_mode *mode)
{
  register int foo = (mode->request.graphtype == GT_TEXT) 
  			   ? mode->request.textgrid.x : 8;

  #define ADJUST(X) mode->x.X -= (mode->x.X % foo)

  ADJUST(width);
  ADJUST(blankstart);
  ADJUST(syncstart);
  ADJUST(syncend);
  ADJUST(blankend);
  if (mode->x.total % foo)
   mode->x.total += foo - (mode->x.total % foo);	

  if (((mode->x.syncend / foo) & 0x3F) == ((mode->x.syncstart / foo) & 0x3F))
    mode->x.syncend -= foo;
  if (((mode->x.blankend / foo) & 0x7F) == ((mode->x.blankstart / foo) & 0x7F))
    mode->x.blankend -= foo;

  if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
    mode->y.syncend--;
  if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
    mode->y.blankend--;

  #undef ADJUST
}

static void
set_origin(unsigned int offset)
{
  IS_CRITICAL;

  ENTER_CRITICAL;
     outbCRT(offset >> 8, 12);
     outbCRT(offset, 13);
  LEAVE_CRITICAL;
}

static void
show_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  IS_CRITICAL;

  ENTER_CRITICAL;
    cursor->state |= PF_SHOWN;
    outbCRT(offset >> 8, 14);
    outbCRT(offset, 15);
  LEAVE_CRITICAL;
}

static void
hide_cursor(struct kgi_pointer *cursor, unsigned int offset)
{
  IS_CRITICAL;

  if (!(cursor->state & PF_SHOWN))
    return;

  ENTER_CRITICAL;
    cursor->state &= ~PF_SHOWN;
    outbCRT(0xFF, 14);
    outbCRT(0xFF, 15);
  LEAVE_CRITICAL;
}

int
kgi_chipset_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int Bpp = 1, mul = mode->lclk.mul, div = mode->lclk.div;

  if (mode->flags & FL_MON_INTERLACE)
    {
      TRACE(printk("%s: interlaced modes not supported (yet).\n",__FILE__));
      return -ECHIPSET; 
    }

  switch(mode->request.graphtype)
   {
     case GT_32BIT:	Bpp +=4;	/* (4+2+1+1)/2 = 4 Bpp	*/
     case GT_15BIT:
     case GT_16BIT:	Bpp +=2;	/*  (2+1+1)/2 =   2 Bpp	*/
     case GT_8BIT:	Bpp +=1;	/*    (1+1)/2 =   1 Bpp	*/
     case GT_4BIT:			/*        1/2 = 0.5 Bpp	*/
     case GT_TEXT:	break;
     default:
	TRACE(printk("%s: graphtype %i not supported.\n", __FILE__, \
		     mode->request.graphtype));
	return -ECHIPSET;
   }

  switch (mode->request.graphtype)
   {
     case GT_TEXT:
	if (cmd == CMD_PROPOSE)
	  {
	    mul = mode->lclk.mul = mode->clk.mul = 1;
	    div = mode->lclk.div = mode->clk.div = 1;
	  }			/* no break; fall through	*/
     case GT_4BIT:
     case GT_8BIT:
	if ((mul != 1) || !((div == 1) || (div == 2)))
	  return -ECHIPSET;
	break;

     case GT_15BIT:
     case GT_16BIT:
	if ((div != 1) || !((mul == 1) || (mul == 1)))
	  return -ECHIPSET;
	break;

     case GT_24BIT:
	if ((mul != 3) || !((div == 1) || (div == 2)))
	  return -ECHIPSET;
	break;

     case GT_32BIT:
	if ((mul != 2) || (div != 1))
	  return -ECHIPSET;
	break;

     default:
	INTERNALERROR;
   }

#if 0
  if ((mode->clk.mul != mul) || (mode->clk.div != div)) 
    {
      TRACE(printk("%s:%i: lclk != clk divider\n",__FILE__,__LINE__));
      return -ECHIPSET;
    }
#endif

  #define XVIRTUAL (mode->request.virtual.x)
  #define YVIRTUAL (mode->request.virtual.y)

  switch (cmd)
   {
     case CMD_PROPOSE:

	if (mode->request.graphtype == GT_TEXT)
	  {
	    mode->magnify.x = mode->magnify.y = 1;
	    if (   (mode->request.textgrid.x < 7) 
	        || (9 < mode->request.textgrid.x))
	      {
	        TRACE(printk("%s: invalid char width (%i).\n",__FILE__,
		      mode->request.textgrid.x));
		return -EINVAL;
	      }

	    if ((2 * (XVIRTUAL / mode->request.textgrid.x)
		   * (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len)
	      {
		TRACE(printk("%s: not enough text memory.\n",__FILE__));
		return -ENOMEM;
	      }
	    
	    mode->request.visible.x -= 
		(mode->request.visible.x % mode->request.textgrid.x);
	    mode->request.visible.y -= 
		(mode->request.visible.y % mode->request.textgrid.y);
	  }
         else	/* mode->request.graphtype == GT_TEXT */
	  {
	    mode->magnify.x = mode->magnify.y = 1;

	    if (XVIRTUAL <=  640) XVIRTUAL =  640; else
	    if (XVIRTUAL <=  800) XVIRTUAL =  800; else
	    if (XVIRTUAL <= 1024) XVIRTUAL = 1024; else
	    if (XVIRTUAL <= 1152) XVIRTUAL = 1152; else
	    if (XVIRTUAL <= 1280) XVIRTUAL = 1280; else
	    if (XVIRTUAL <= 1600) XVIRTUAL = 1600; else
	    if (XVIRTUAL <= 2048) XVIRTUAL = 2048;
              else
	       {
	         TRACE(printk("%s: %i horizontal pixels are too many.\n",\
			      __FILE__, XVIRTUAL));
	         return -ECHIPSET;
	       }
	
	    if ((XVIRTUAL* YVIRTUAL* Bpp / 2) > chipinfo.ramsize)
	      {
	        TRACE(printk(\
		"%s: not enough memory to do virtual %ix%i at %i bpp",\
		__FILE__, XVIRTUAL, YVIRTUAL, Bpp*4));
	        return -ENOMEM;
	      }

	    mode->request.visible.x &= 0x7FFFFFF8;
	    mode->request.visible.y &= 0x7FFFFFF8;
	}

	mode->x.width = mode->request.visible.x;
	mode->y.width = mode->request.visible.y;

	return EOK;

     case CMD_LOWER:

	if (mode->dclk < chipinfo.dclk.min)
	  {
	    TRACE(printk("%s: DCLK too low (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

	mode->dclk = (mode->dclk > chipinfo.dclk.max)
	 		  ?  chipinfo.dclk.max : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_RAISE:

	if (mode->dclk > chipinfo.dclk.max)
	  {
	    TRACE(printk("%s: DCLK too high (%i).\n",__FILE__, mode->dclk));
	    return -ECHIPSET;
          }

	mode->dclk = (mode->dclk < chipinfo.dclk.min)
			?  chipinfo.dclk.min : mode->dclk;

	adjust_timing(mode);
	return EOK;

     case CMD_CHECK:
	if (mode->request.graphtype == GT_TEXT)
	  {
	     mode->set_origin  = set_origin;
	     mode->cursor_hide = hide_cursor;
	     mode->cursor_show = show_cursor;
	     mode->flags |= FL_CHIP_COLORTEXT;
	     if (   (mode->magnify.x != 1) || (mode->magnify.y != 1) /*driver*/
		 || (mode->request.textgrid.x < 7) /* hardware */
		 || (9 < mode->request.textgrid.x) /* hardware */
		 || (mode->request.textgrid.y < 1) /* hardware */
		 || (mode->request.textgrid.y > 32) /* hardware */
		 || ((2 * (XVIRTUAL / mode->request.textgrid.x)
			* (YVIRTUAL / mode->request.textgrid.y)) > fb_text_len)
		 || (mode->dclk > chipinfo.dclk.max) /* hardware */
		 || (mode->dclk < chipinfo.dclk.min)) /* hardware */
		return -ECHIPSET;
	  }
	 else
	  {
	    switch (XVIRTUAL)	
	     {
	       case 640:  case 800:  case 1024:	case 1152:
	       case 1280: case 1600: case 2048:
	     	  break;
	       default:
	  	  return -ECHIPSET;
	     }

  	    if (   ((XVIRTUAL * YVIRTUAL * Bpp / 2) > chipinfo.ramsize)
	        || (mode->magnify.x != 1) || (mode->magnify.y != 1)/*!!! temp */
	        || (mode->dclk < chipinfo.dclk.min) 
		|| (mode->clk.mul != 1) || (mode->clk.div != 1)
	        || (chipinfo.dclk.max < mode->dclk))
	      return -ECHIPSET;
	  }

	#define TM(X) (   ((unsigned) chipinfo.maxdots.x < mode->x.X)\
		       || ((unsigned) chipinfo.maxdots.y < mode->y.X))
	if (   TM(width)   || TM(blankstart) || TM(syncstart)
	    || TM(syncend) || TM(blankend)   || TM(total))
	  return -ECHIPSET;
 	#undef TM

	adjust_timing(mode);

	{				/* calculate the M parameter	*/
	  int m, clock;			/* (was taken from XFree86)	*/

	  clock = mode->dclk*Bpp/2000;
	  if (chipinfo.ramsize < 2048*1024)
	    clock += clock;
	  if (clock > 230000)
	    m = 0;
	   else
	    m = (int)(5384.18/(clock/1000.0+39)-21.1543);
	  if (chipinfo.ramsize < 2048*1024)
	    clock /= 2;

	  if (m < 0)
	    m = 0;
	   else
	    if (m > 31)
	      m = 31;
	  mode->chipset = m << 3;
	} 

	return EOK;

     default:
	INTERNALERROR;
	return -ECHIPSET;
   }

  #undef XVIRTUAL 
  #undef YVIRTUAL
}

void
kgi_chipset_set_mode (struct kgi_mode *mode)
{
  IS_CRITICAL;
  struct statetable setup = default_state;
  register int foo;
  int DCLKdiv;

  /*\
  |*|	Setup bits per pixel and logical screen width
  \*/

  switch(mode->request.graphtype)
   {
     case GT_TEXT:
       	foo = 2 * mode->request.virtual.x / mode->request.textgrid.x;
	foo /= 4; /* due to the addressing mode used. */ 
        break;
     case GT_4BIT:
	setup.CRT[0x50] |= CR50_4BPP;
	foo = mode->request.virtual.x >> 4;
	break;
     case GT_8BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_8BPP;
	foo = mode->request.virtual.x >> 3;
	break;
     case GT_15BIT:
     case GT_16BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_16BPP;
/* !!!	setup.CRT[0x67] |= CR67_NO_INVERT_VCLK;	*/
	foo = mode->request.virtual.x >> 2;
	break;
     case GT_32BIT:
	setup.CRT[0x3A] |= CR3A_NO4BPPMODE;
	setup.CRT[0x50] |= CR50_32BPP;
	foo = mode->request.virtual.x >> 1;
	break;

     default:
	INTERNALERROR;
	return;
   }
   setup.CRT[CRT_LOGICALWIDTH]  |= foo & 0xFF;
   setup.CRT[0x51] |= (foo & 0x100) ? CR51_LOGICALWIDTH_B8 : 0;
   setup.CRT[0x51] |= (foo & 0x200) ? CR51_LOGICALWIDTH_B9 : 0;

   if(mode->request.graphtype == GT_TEXT)
     {
       setup.CRT[0x17] =   CR17_CGA_ADDR_WRAP | CR17_WORDMODE
                         | CR17_ENABLE_SYNC   | CR17_CGA_BANKING
                         | CR17_HGC_BANKING;

       setup.CRT[0x31] |= CR31_HISPEEDTXT;

       setup.SEQ[0x02] = SR02_PLANE0 | SR02_PLANE1;

       setup.GRC[0x05] |= GR5_ODD_EVEN;
       setup.GRC[0x06] |= GR6_MAP_B8_32 | GR6_CHAIN_OE;

       /* setup.CRT[0x34] |= CR34_ENB_FIFO;	???*/
	
       switch((DCLKdiv = mode->request.textgrid.x))
	{
	  case 7: setup.CRT[0x32] |= CR32_7DOT_CHCLK; break;
	  case 8: setup.SEQ[0x01] |= SR01_8DOT_CHCLK; break;
	  case 9: break;
	  default:
	     INTERNALERROR;
	     return;
	}
       setup.CRT[0x09] |= mode->request.textgrid.y - 1;
       setup.ATC[0x13] = (mode->request.textgrid.x == 9) ? 0x08 : 0x00;
     }
    else /* mode->request.graphtype == GT_TEXT */
     {
       DCLKdiv = 8;
       
       for (foo = 0; foo < 16; foo++) setup.ATC[foo] = foo;

       setup.SEQ[0x01] |= SR01_8DOT_CHCLK;
       setup.SEQ[0x02] |= SR02_PLANE_MASK;/*??? enable all planes-neccessary ?*/ 
       setup.ATC[0x10] |= AR10_GRAPHICS;
       setup.GRC[0x05] |= GR5_8BIT_COLOR; /*??? neccessary, I think not */
       setup.GRC[0x06] |= GR6_GRAPHMODE | GR6_MAP_A0_64 /*??? neccessary ?*/;
       setup.CRT[0x31] |= CR31_ENHANCED;
       setup.CRT[0x33] |= CR33_NOBORDER;
       setup.CRT[0x58] |= CR58_LAW_ENABLE;


       setup.CRT[0x09] |= /* !!! TM->ymagnify-1 & CR09_CHARHEIGHT_MASK; */ 0;
       setup.CRT[0x09] |= (mode->flags & FL_CHIP_DOUBLESCAN) 
				? CR09_DOUBLESCAN : 0;

       switch(mode->request.virtual.x)
        { 
          case  640: setup.CRT[0x50] |= CR50_640; break;
          case  800: setup.CRT[0x50] |= CR50_800; break;
          case 1024: setup.CRT[0x50] |= CR50_1024; break;
          case 1152: setup.CRT[0x50] |= CR50_1152; break;
          case 1280: setup.CRT[0x50] |= CR50_1280; break;
          case 1600: setup.CRT[0x50] |= CR50_1600; break;
          case 2048: setup.CRT[0x50] |= CR50_2048;
		     setup.CRT[0x31] |= CR31_2PAGE_SCRN; 
		     break;
          default: INTERNALERROR;
        }

       foo = 0;
       switch(mode->request.graphtype)
        {
          case GT_TEXT:
          case GT_4BIT:
          case GT_8BIT:
		foo = (mode->lclk.div == 1) ? CR67_DAC_MODE0 : CR67_DAC_MODE8;
		break;

          case GT_15BIT:
		foo = (mode->lclk.mul == 2) ? CR67_DAC_MODE1 : CR67_DAC_MODE9;
		break;

          case GT_16BIT:
		foo = (mode->lclk.mul == 2) ? CR67_DAC_MODE3 : CR67_DAC_MODEA;
		break;


          case GT_24BIT:
/* ??? */	foo = (mode->lclk.div == 1) ? CR67_DAC_MODE2 : CR67_DAC_MODEB;
		break;

          case GT_32BIT:
		foo = CR67_DAC_MODEB;
		break;

          default:
		INTERNALERROR;
		return;
        }
       setup.CRT[0x67] |= foo;
     }

  /*\
  |*|	setup horizontal timing
  \*/
 
   foo = (mode->x.width / DCLKdiv) - 1;		/* horizontal display end */
     setup.CRT[0x01] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HDISPLAYEND_B8 : 0;

   foo = (mode->x.blankstart / DCLKdiv);	/* horizontal blank start */
     setup.CRT[0x02] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HBLANKSTART_B8 : 0;

   foo = (mode->x.syncstart / DCLKdiv);		/* horizontal sync start  */
     setup.CRT[0x04] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HSYNCSTART_B8 : 0;

   foo = (mode->x.syncend / DCLKdiv);		/* horizontal sync end    */
     setup.CRT[0x05] |= foo & CR05_HSE_MASK;
     setup.CRT[0x5D] |= ((foo- (mode->x.syncstart/DCLKdiv)) > 0x20)
				 ? CR5D_HSYNCEND_B5 : 0;


   foo = (mode->x.blankend / DCLKdiv); 		/* horizontal blank end   */
     setup.CRT[0x03] |= foo & CR03_HBE_MASK;
     setup.CRT[0x05] |= (foo & 0x20) ? CR05_HBLANKEND_B5 : 0;
     setup.CRT[0x5D] |= ((foo - (mode->x.blankstart/DCLKdiv)) > 0x40)
				 ? CR5D_HBLANKEND_B6 : 0;
  
   foo = (mode->x.total / DCLKdiv) - 5;		/* horizontal total       */
     setup.CRT[0x00] |= foo & 0xFF;
     setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_HTOTAL_B8 : 0;

  /*\
  |*|	setup vertical timing
  \*/
 
   foo = mode->y.width - 1;			/* vertical display end   */
     setup.CRT[0x12] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VDISPLAYEND_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VDISPLAYEND_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VDISPLAYEND_B10 : 0;

   foo = mode->y.blankstart - 1;		/* vertical blank start   */
     setup.CRT[0x15] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VBLANKSTART_B8  : 0;
     setup.CRT[0x09] |= (foo & 0x200) ? CR09_VBLANKSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VBLANKSTART_B10 : 0;
    
   foo = mode->y.syncstart;			/* vertical sync start    */
     setup.CRT[0x10] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VSYNCSTART_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VSYNCSTART_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VSYNCSTART_B10 : 0;

   foo = mode->y.syncend;			/* vertical sync end      */
     setup.CRT[0x11] |= foo & CR11_VSYNCEND_MASK;
 
   foo = mode->y.blankend + 1;			/* end vertical blank     */
     setup.CRT[0x16] |= foo & 0xFF;
  
   foo = mode->y.total - 2;			/* vertical total         */
     setup.CRT[0x06] |= foo & 0xFF;
     setup.CRT[0x07] |= (foo & 0x100) ? CR07_VTOTAL_B8  : 0;
     setup.CRT[0x07] |= (foo & 0x200) ? CR07_VTOTAL_B9  : 0;
     setup.CRT[0x5E] |= (foo & 0x400) ? CR5E_VTOTAL_B10 : 0;
 
   setup.CRT[CRT_INTERLACESTART]|= mode->y.total >> 1; /* ??? wrong ? (FIXME) */
   if(mode->flags & FL_MON_INTERLACE)
    {
       printk("%s:%i: no interlaced modes (yet).\n",__FILE__,__LINE__);
       setup.CRT[0x42] |= CR42_INTERLACED;
       INTERNALERROR;
       return; /* !!! temporary */
    }

  setup.MISC |= (mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC;
  setup.MISC |= (mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC;

#if 0
  for (foo = 0; !((1 << foo) & mode->lclk.div); foo++);
  setup.CRT[0x66] |= foo & CR66_VCLK_DIV_MASK;
#endif

  /*\
  |*|	Setup the FIFO control parameters.	!!! this needs some tuning!	
  \*/

  foo = (mode->x.total + mode->x.syncstart)/2 / DCLKdiv;
  setup.CRT[0x3B] |= foo & 0xFF;	/* start display FIFO fetch	*/
  setup.CRT[0x5D] |= (foo & 0x100) ? CR5D_FIFOSTART_B8 : 0;

  if (mode->request.graphtype == GT_TEXT)
    {
      setup.CRT[0x54] = 8;
      setup.CRT[0x60] = 7;
      setup.CRT[0x62] = 84;
    }
   else
    {
      setup.CRT[0x54] = mode->chipset;	/* M parameter	*/
      setup.CRT[0x60] = 0xFF;		/* N parameter	*/

      foo = mode->request.visible.x;
      switch (mode->request.graphtype) 
       {
         case GT_4BIT:	foo /= 2; break;

         case GT_8BIT:	break;

         case GT_15BIT:
         case GT_16BIT:	foo *= 2; break;

         case GT_32BIT:	foo *= 4; break;
       }
      foo /= (chipinfo.ramsize < 2048*1024) ? 4 : 8;
      setup.CRT[0x61] |= ((foo >> 8) & CR61_L_DATAMASK) | CR61_L_ENABLE;
      setup.CRT[0x62] = foo & 0xFF;
    }
  /*\
  |*|	Ok. We'll do the programming now.
  \*/

  ENTER_CRITICAL;
    if (current_mode.graphtype != GT_TEXT)
      {
        foo = 10000;
        while((inb(ECR_GP_STATUS) & 0xF2FF) && foo)
          foo--;
      }
    outb(0, ECR_ADVFUNC_CNTL);	/* disable accelerator engine */

    outbCRT(inb(CRT_EXTSYNC1) | CR56_SYNC_OFF, CRT_EXTSYNC1);
    kgi_monitor_set_mode(mode);
    kgi_ramdac_set_mode(mode);
    kgi_clock_set_mode(mode);
    s3_unlock();
    restore_state(&setup);
  LEAVE_CRITICAL;

  if (mode->request.graphtype == GT_TEXT)
    {						/* set text fonts	*/
      getsetfont(SETFONT, 0, mode->font0);
      if (mode->font1)
        getsetfont(SETFONT, 1, mode->font1);
       else
        getsetfont(SETFONT, 1, mode->font0);
    }
   else
    {
	/* !!! temporary, this will be moved to an extra source file. */

	/* Reset the Accelerator Engine */
	outw(((mode->request.graphtype == GT_4BIT) ? EC01_4BPP : EC01_ENH_ENB)
              | EC01_LAW_ENB, ECR_ADVFUNC_CNTL);
	outw(EC00_RESET,  ECR_SUBSYS_CNTL);
	inw_p(ECR_SUBSYS_STAT);	/* wait a bit */
	outw(EC00_ENABLE, ECR_SUBSYS_CNTL);
	inw_p(ECR_SUBSYS_STAT); /* wait a bit */

	outw(MF_MULT_MISC, ECR_MULTIFUNC); /* 16 bit I/O registers */

        /* Now set some basic accelerator registers */
	wait_fifo(6);
	outw(SRC_FOREGROUND | MIX_SRC, ECR_FOREGR_MIX);
	outw(SRC_BACKGROUND | MIX_SRC, ECR_BACKGR_MIX); /* Use color directly */
	outw(MF_T_CLIP | 0, ECR_MULTIFUNC);		/* clip virtual area  */
	outw(MF_L_CLIP | 0, ECR_MULTIFUNC);
	outw(MF_R_CLIP | (mode->request.virtual.x-1), ECR_MULTIFUNC);
	outw(MF_B_CLIP | (mode->request.virtual.y-1), ECR_MULTIFUNC);

	wait_fifo(5);
	outw(~0, ECR_WRITE_MASK);			/* Enable all planes */
	outw(~0, ECR_WRITE_MASK);			/* Enable all planes */
	outw(MF_PIX_CONTROL | MFA_SRC_FOREGR_MIX, ECR_MULTIFUNC);

	/* !!! reset other registers too. */
	memset(fb_graphics, 0, fb_graph_len);
    }

  #ifdef DEBUG_IO 
   dump_state(&setup); printk("\n");
  #endif
}

int
kgi_chipset_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  return -ECHIPSET;
}

#ifdef NOACCEL
#include "s3_noaccel.inc"
#else
#include "s3_accel.inc"
#endif

#include "linearmap.inc"
