/*\
|*| ---------------------------------------------------------------------------
|*|	ATT20c505 / ATT20c504 ramdac driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#define	DAC_VERSION	"0.12"
#undef	DEBUG


#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif
#define DAC_PRECISION	8	/* we can do 8 bit R,G,B resolution	*/

static unsigned char CR_0,CR_1,CR_2,CR_3;

#define __NO_VERSION__
#include <linux/module.h>
#include <linux/version.h>
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/types.h>
#include <linux/event.h>
#include <linux/sched.h>

#include <ggi/graphdev.h>
#include <ggi/ATT/20c505.h>
#include <ggi/IBM/clut.inc>

#define PCLK DAC_C2_PCLK0     /* or DAC_C2_PCLK1 ? (doesn't matter on 20SV) */

#if !((MONITORVERSION == 01) && (MODEVERSION == 03) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif


enum RAMDAC_types 
 {
	ATT20C504_85	= 0,
	ATT20C504_11	= 1,

	ATT20C505_11	= 2,
	ATT20C505_13	= 3,
	ATT20C505_17	= 4,

	BT485_11	= 5,
	BT485_13	= 6,

	DAC_UNKNOWN	= -1
 };

#define ATT20C504 ATT20C504_85
#define ATT20C505 ATT20C505_11
#define	BT485 BT485_11

#define	SUPPORTED (GM_TEXT | GM_1BIT | GM_4BIT | GM_8BIT \
		  | GM_15BIT | GM_16BIT | GM_32BIT | GM_4BIT_PLANES )

static struct kgi_ramdac att_params[7] =
      {
       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C504-85",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0,  85000000 },	/* normal mode		*/
	  {       0,         0 },	/* 2xclk mode		*/

	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0,  85000000 },	/* 	GT_1BIT		*/
	  {	  0,  85000000 },	/* 	GT_4BIT		*/
	  {       0,  85000000 },	/* 	GT_8BIT		*/	
	  {	  0,  85000000 },	/* 	GT_15BIT	*/
	  { 	  0,  85000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 },	/* 	GT_32BIT	*/
	  {	  0,  85000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C504-11",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA,	/* flags		*/
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* limit in normal mode	*/
	  {       0,         0 },	/* limit in 2xclk mode	*/
 
	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0, 110000000 },	/* 	GT_1BIT		*/
	  {	  0, 110000000 },	/* 	GT_4BIT		*/
	  {       0, 110000000 },	/* 	GT_8BIT		*/	
	  {	  0, 110000000 },	/* 	GT_15BIT	*/
	  { 	  0, 110000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 },	/* 	GT_32BIT	*/
	  {	  0, 110000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-11",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  55000000 },	/* limit in 2xclk mode	*/

	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0, 110000000 },	/* 	GT_1BIT		*/
	  {	  0, 110000000 },	/* 	GT_4BIT		*/
	  {       0, 110000000 },	/* 	GT_8BIT		*/	
	  {	  0, 110000000 },	/* 	GT_15BIT	*/
	  { 	  0, 110000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 },	/* 	GT_32BIT	*/
	  {	  0, 110000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-13",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  67500000 },	/* 2xclk mode		*/

	 {{	  0,  85000000 },	/* 	GT_TEXT		*/
	  {       0, 135000000 },	/* 	GT_1BIT		*/
	  {	  0, 135000000 },	/* 	GT_4BIT		*/
	  {       0, 135000000 },	/* 	GT_8BIT		*/	
	  {	  0, 135000000 },	/* 	GT_15BIT	*/
	  { 	  0, 135000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  85000000 },	/* 	GT_32BIT	*/
	  {	  0, 135000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "AT&T Microelectronics",	/* manufacturer		*/
	 "ATT 20C505-17",		/* model		*/
	 SUPPORTED,			/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  { 	  0, 110000000 },	/* normal mode		*/
	  {45000000,  85000000 },	/* 2xclk mode		*/

	 {{	  0,  85000000 },	/* 	GT_TEXT		*/
	  {       0, 135000000 },	/* 	GT_1BIT		*/
	  {	  0, 175000000 },	/* 	GT_4BIT		*/
	  {       0, 175000000 },	/* 	GT_8BIT		*/	
	  {	  0, 175000000 },	/* 	GT_15BIT	*/
	  { 	  0, 175000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  85000000 },	/* 	GT_32BIT	*/
	  {	  0, 175000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "Brooktree Corporation",	/* manufacturer		*/
	 "Bt 485-11",			/* model		*/
	 SUPPORTED & ~GM_1BIT,		/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA |	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  55000000 },	/* limit in 2xclk mode	*/

	 {{	  0,  75000000 },	/* 	GT_TEXT		*/
	  {       0, 110000000 },	/* 	GT_1BIT		*/
	  {	  0, 110000000 },	/* 	GT_4BIT		*/
	  {       0, 110000000 },	/* 	GT_8BIT		*/	
	  {	  0, 110000000 },	/* 	GT_15BIT	*/
	  { 	  0, 110000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  75000000 },	/* 	GT_32BIT	*/
	  {	  0, 110000000 }}	/* 	GT_4BIT_PLANES	*/
       },

       { "Brooktree Corporation",	/* manufacturer		*/
	 "Bt 485-13",			/* model		*/
	 SUPPORTED & ~GM_1BIT,		/* supported modes	*/
	 FL_DAC_8BIT | FL_DAC_GAMMA | 	/* flags		*/
	 FL_DAC_2XCLOCK,
	 { 0x0FFF, 0x0FFF },		/* maximum resolution	*/

	 DAC_8BIT_PB | 			/* device type		*/
	 DAC_32BIT_PB | DAC_VRAM,
	  {       0, 110000000 },	/* normal mode		*/
	  {45000000,  67500000 },	/* 2xclk mode		*/

	 {{	  0,  85000000 },	/* 	GT_TEXT		*/
	  {       0, 135000000 },	/* 	GT_1BIT		*/
	  {	  0, 135000000 },	/* 	GT_4BIT		*/
	  {       0, 135000000 },	/* 	GT_8BIT		*/	
	  {	  0, 135000000 },	/* 	GT_15BIT	*/
	  { 	  0, 135000000 },	/* 	GT_16BIT	*/
	  {       0,         0 },	/* 	GT_24BIT	*/
	  {	  0,  85000000 },	/* 	GT_32BIT	*/
	  {	  0, 135000000 }}	/* 	GT_4BIT_PLANES	*/
       }
      };

static struct kgi_ramdac *dac_params = NULL;


/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
   unsigned char status = inbDAC(DAC_STATUSREG);
   
   TRACE(printk("ATT/Brooktree driver v"DAC_VERSION"\n"));

   switch(status & DAC_ST_ID_MASK)
    {
      case DAC_BT485:
	dac_params = att_params + BT485;
	break;

      case DAC_ATT20C504:
	dac_params = att_params + ATT20C504;
	break;

      case DAC_ATT20C505:
	dac_params = att_params + ATT20C505;
	break;

      case DAC_ATT20C506:
	printk("%s: don't know how to handle ATT20C506 device.\n",\
		__FILE__);
	return -ERAMDAC;

      default:
	printk("%s: unknown chip ID (%.2x).\n",__FILE__,\
	        status);
	/*	 status & DAC_ST_ID_MASK); */
	return -ERAMDAC;
    }
  TRACE(printk("  %s %s detected.\n", dac_params->manufact, dac_params->model));
  return EOK;
}

void
kgi_ramdac_done(void)
{
  TRACE(printk("%s %s driver removed.\n", dac_params->manufact,\
        dac_params->model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return dac_params; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int div;
  int newclock = mode->dclk;
  
  switch(mode->request.graphtype)
   {
     case GT_TEXT:	div =  1; break;
     case GT_1BIT:	div = 32; break;
     case GT_4BIT_PLANES:
     case GT_4BIT:	div =  8; break;
     case GT_8BIT:	div =  4; break;
  
     case GT_15BIT:
     case GT_16BIT:	div =  2; break;
  
     case GT_32BIT:	div =  1; break;
  
     case GT_24BIT:	return -ERAMDAC;
     default:
  	TRACE(printk("%s: could not handle graphtype %s\n",\
  	     __FILE__, GRAPHMASK(mode->request.graphtype)));
 	return -ERAMDAC;
   }

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = mode->clk.div = 1;
        mode->lclk.div = div;
	TRACE(printk("%s: %i:%i multiplexing mode.\n", \
		     __FILE__, mode->lclk.div, mode->lclk.mul));

	mode->flags &= ~FL_DAC_2XCLOCK;
	mode->flags |= FL_DAC_8BIT;
	if((mode->request.graphtype == GT_32BIT) || 
	   (mode->request.graphtype == GT_15BIT) ||
	   (mode->request.graphtype == GT_16BIT))
	 mode->flags |= FL_DAC_GAMMA;

	return EOK;

     case CMD_LOWER:
	if (   (newclock < dac_params->clock[mode->request.graphtype].min)
	    || (newclock < dac_params->normal.min))
	  {
	    TRACE(printk("%s: DCLK too low (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (dac_params->clock[mode->request.graphtype].max < newclock)
    	  newclock = dac_params->clock[mode->request.graphtype].max;

	if (dac_params->normal.max < newclock)
	  {
	    TRACE(printk("%s: clock doubling mode not implemented yet.\n",\
			 __FILE__));
	    newclock = dac_params->normal.max;
	  }

	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > dac_params->clock[mode->request.graphtype].max)
            || (newclock > dac_params->normal.max))
	  {
	    TRACE(printk("%s: DCLK too high (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (newclock < dac_params->clock[mode->request.graphtype].min)
	  newclock = dac_params->clock[mode->request.graphtype].min;

	if (newclock < dac_params->normal.min)
	  newclock = dac_params->normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
	if (   (mode->lclk.mul != 1) || (mode->lclk.div != div)
	    || (mode->clk.mul  != 1) || (mode->clk.div !=    1))
	  return -ERAMDAC;

	if (   (mode->dclk < dac_params->normal.min)
	    || (mode->dclk < dac_params->clock[mode->request.graphtype].min))
	  return -ERAMDAC;

	if (   (mode->dclk > dac_params->normal.max)
	    || (mode->dclk > dac_params->clock[mode->request.graphtype].max))
	  return -ERAMDAC;

	return EOK;

    default:
	INTERNALERROR;
	return -ERAMDAC;
  }
}

void
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  struct kgi_monitor *monitor;

  /*\
  |*|	setup CR_0
  \*/
     CR_0 = DAC_C0_8BIT;		/* we'll always use 8 bit DAC */
     monitor = kgi_monitor_get_info();
     if (monitor->synctype & SYNC_ON_RED)     CR_0 |= DAC_C0_SYNC_ON_RED;
     if (monitor->synctype & SYNC_ON_GREEN)   CR_0 |= DAC_C0_SYNC_ON_GREEN;
     if (monitor->synctype & SYNC_ON_BLUE)    CR_0 |= DAC_C0_SYNC_ON_BLUE;
     if (monitor->synctype & BLANK_PEDESTRAL) CR_0 |= DAC_C0_BLANK_LEVEL;

  /*\
  |*|	setup CR_1
  \*/
     switch(mode->request.graphtype)
      {
        case GT_TEXT:	CR_1 = 0;		break;
        case GT_1BIT:	CR_1 = DAC_C1_1BPP;	break;
	case GT_4BIT_PLANES:
	case GT_4BIT:	CR_1 = DAC_C1_4BPP;	break;
	case GT_8BIT:	CR_1 = DAC_C1_8BPP;	break;
	case GT_15BIT:	CR_1 = DAC_C1_15BPP;	break;
	case GT_16BIT:	CR_1 = DAC_C1_16BPP;	break;
	case GT_32BIT:	CR_1 = DAC_C1_24BPP;	break;
	default:
	  INTERNALERROR;	/* for safety, and to trick gcc */
      };
     TRACE(printk("%s: graphtype = %s.\n", \
	    __FILE__, GRAPHMASK(mode->request.graphtype)));

  /*   CR_1 |= DAC_C1_CLUT_BYPASS;  turn on for true-color debugging only. */

  /*\
  |*|	setup CR_2
  \*/
      CR_2 =  ((mode->request.graphtype == GT_TEXT) ? 0:  DAC_C2_ENHANCED_MODE)
           | ((mode->flags & FL_MON_INTERLACE)    ? DAC_C2_INTERLACED : 0)
           | PCLK;

  /*\
  |*|	setup CR_3
  \*/
     CR_3 = 0;		/* we have no clock doubling yet ... */

  /*\
  |*|	program CR_3,CR_2,CR_1,CR_0 (CR_3 looks a bit strange, but is correct ;-)
  \*/
     outbDAC(CR_0 | DAC_C0_EXT_UNLOCK | DAC_C0_SLEEP_MODE, DAC_CONTROL0);
     outbDAC(DAC_PW_CR3_INDEX, DAC_PW_INDEX);
     outbDAC(CR_3, DAC_STATUSREG);

     outbDAC(CR_2, DAC_CONTROL2);
     outbDAC(CR_1, DAC_CONTROL1);
     outbDAC(CR_0, DAC_CONTROL0);

     outbDAC(0xFF, DAC_PIXEL_MASK);

  set_CLUT_entries(mode->palette);	/* set clut of mode */
}

int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
  struct kgi_state *gr = file->private_data;

  switch(cmd)
  { case RAMDAC_SETCLUT:
	{  ggi_palette *h = (void*) _io_buf;
	   int first,cnt;

	   first=h->first;cnt=h->cnt;
           if (first+cnt > 256) return -ENBOGUS;
           if (0==verify_area (VERIFY_READ, (void *) h->data, sizeof(ggi_color)*cnt))
           { memcpy_fromfs (&(gr->color[first]), 
           		    (void *) h->data, sizeof(ggi_color)*cnt); }
	   set_CLUT_entries(gr->mode.palette);		/* set clut of mode */
	} 
	return 0;
    case RAMDAC_GETCLUT:
	{  ggi_palette *h = (void*) _io_buf;
	   int first,cnt;
	   _out_buf=_io_buf;

	   first=h->first;cnt=h->cnt;
           if (first+cnt > 256) return -ENBOGUS;
           if (0==verify_area (VERIFY_WRITE, (void *) h->data, sizeof(ggi_color)*cnt))
           { memcpy_tofs ((void *) h->data,
           		  &(gr->color[first]), sizeof(ggi_color)*cnt); }
	}
	return 0;
  }
  return -ENODRVSUP_ALWAYS_CANT;	/* This driver can't do this. */
}
