/*\
|*| ---------------------------------------------------------------------------
|*|	VGA ramdac driver
|*| ---------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Steffen Seeger, [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

#define	MAINTAINER	"Steffen Seeger, [seeger@physik.tu-chemnitz.de]"
#define	DAC_VERSION	"0.05 ALPHA"

#ifdef DEBUG
#define	NEED_GRAPHTYPE_SYMS
#endif

#include <ggi/graphdev.h>

#define	DAC_PRECISION	6	/* we can do 6 bit per RGB value	*/

#include <ggi/IBM/vga_dac.h>
#include <ggi/IBM/clut.inc>

#if !((MONITORVERSION == 01) && (MODEVERSION == 03) && (RAMDACVERSION == 01))
  #warning kgi datastructures have changed, check this code !!
#endif

static struct kgi_ramdac 
ramdac = {	"IBM",				/* manufacturer		*/
	 	"VGA",				/* model		*/
 		GM_TEXT | GM_4BIT | GM_8BIT
 		| GM_4BIT_PLANES,		/* supported modes	*/
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/

		 DAC_8BIT_PB,  			/* device type		*/
		  {       0,  32000000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  32000000 },	/* 	GT_TEXT		*/
		  {       0,         0 },	/* 	GT_1BIT		*/
		  {	  0,  32000000 },	/* 	GT_4BIT		*/
		  {       0,  32000000 },	/* 	GT_8BIT		*/	
		  {	  0,         0 },	/* 	GT_15BIT	*/
		  { 	  0,         0 },	/* 	GT_16BIT	*/
		  {       0,         0 },	/* 	GT_24BIT	*/
		  {	  0,         0 },	/* 	GT_32BIT	*/
		  {       0,  32000000 }}	/* 	GT_4BIT_PLANES	*/	
         };

/*\
|*|	GGI interface
\*/

int
kgi_ramdac_init(void)
{
  TRACE(printk("%s %s dac driver v"DAC_VERSION"\n", \
        ramdac.manufact, ramdac.model));
  return EOK;
}

void
kgi_ramdac_done(void)
{
  TRACE(printk("%s %s driver removed.\n", ramdac.manufact, ramdac.model));
}

struct kgi_ramdac
*kgi_ramdac_get_info (void)
{
  return &ramdac; 
}

int
kgi_ramdac_check_mode (struct kgi_mode *mode, enum kgi_tm_cmd cmd)
{
  int newclock = mode->dclk;
  
  switch(mode->request.graphtype)
   {
     case GT_TEXT:	break;
     case GT_4BIT_PLANES:	break;
     case GT_4BIT:	break;   /* Really possible ? */
     case GT_8BIT:	break;
  
     default:
  	TRACE(printk("%s: could not handle graphtype %s\n",\
  	     __FILE__, GRAPHMASK(mode->request.graphtype)));
 	return -ERAMDAC;
   }

  switch (cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    TRACE(printk("%s: DCLK too low (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;

	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    TRACE(printk("%s: DCLK too high (%i Hz).\n", __FILE__, newclock));
	    return -ERAMDAC;
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
	if (   (mode->lclk.mul != 1) || (mode->lclk.div != 1)
	    || (mode->clk.mul  != 1) || (mode->clk.div  != 1))
	  return -ERAMDAC;

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -ERAMDAC;

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -ERAMDAC;

	if (kgi_monitor_get_info()->synctype & COMPOSITE_SYNC)
	  {
	    TRACE(printk("%s: composite syncing not supported.\n",__FILE__));
	    return -ERAMDAC;
	  }

	return EOK;

    default:
	INTERNALERROR;
	return -ERAMDAC;
  }
}

void
kgi_ramdac_set_mode(struct kgi_mode *mode)
{
  outbDAC(0xFF, DAC_PIXEL_MASK);
  set_CLUT_entries(mode->palette);	/* set clut of mode */
}

int
kgi_ramdac_ioctl (struct inode *inode, struct file *file, unsigned int cmd,
		  unsigned long arg)
{
    struct kgi_state *gr = file->private_data;

    switch(cmd) { 
	case RAMDAC_SETCLUT: 
         {  
           ggi_palette *h = (void*) _io_buf;
           int first,cnt;

           first=h->first;cnt=h->cnt;
           if (first+cnt > 256)
             return -ENBOGUS;

           if (0==verify_area(VERIFY_READ,(void *)h->data,sizeof(ggi_color)*cnt)) { 
             memcpy_fromfs (&(gr->color[first]),
                            (void *) h->data, sizeof(ggi_color)*cnt); 
           }
           set_CLUT_entries(gr->mode.palette);          /* set clut of mode */
         }
         return 0;
        case RAMDAC_GETCLUT: 
         { 
           ggi_palette *h = (void*) _io_buf;
           int first,cnt;
           _out_buf=_io_buf;

           first=h->first;cnt=h->cnt;
           if (first+cnt > 256)
             return -ENBOGUS;

           if (0==verify_area(VERIFY_WRITE,(void *)h->data,sizeof(ggi_color)*cnt)) { 
             memcpy_tofs ((void *) h->data,
             		  &(gr->color[first]),sizeof(ggi_color)*cnt); 
           }
        }
        return 0;
  }
  return -ENODRVSUP_ALWAYS_CANT;        /* This driver can't do this. */
}
