/*\
|*| --------------------------------------------------------------------------
|*|     GGI general definitons, structures, etc.
|*| --------------------------------------------------------------------------
|*|
|*|	Copyright (C) 1995 by Andreas Beck [becka@hp.rz.uni-duesseldorf.de]
|*|			  and Steffen Seeger [seeger@physik.tu-chemnitz.de]
|*|
|*|	This program is free software; you can redistribute it and/or modify
|*|	it under the terms of the GNU General Public License as published by
|*|	the Free Software Foundation; either version 2, or (at your option)
|*|	any later version.
|*|
|*|	You should have received a copy of the GNU General Public License
|*|	along with this program; see the file COPYING.  If not, write to
|*|	the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
|*|
|*| ---------------------------------------------------------------------------
\*/

/* MAINTAINER	linux-ggi@eskimo.com */

#ifndef	_ggi_h
#define	_ggi_h

#include <ggi/ggi-version.h>

#define	GGIVERSION	002

	typedef struct { int min, max; }	ggi_range;
	typedef struct { int mul, div; }	ggi_ratio;
	typedef struct { int x, y; }		ggi_coord;
	typedef struct { ggi_coord a, b; }	ggi_rect;
	typedef struct { unsigned short r,g,b;}	ggi_color;
	#define GGI_COLOR_PRECISION 10	/* 10 bit per R,G,B value	*/
        typedef struct {   int fg_color;
			   int bg_color;
			   int drawmode; }      ggi_drawinfo;
        enum powersave  { PWR_ON=0,PWR_STANDBY,PWR_SUSPEND,PWR_OFF };
        

	typedef enum
	{	GT_TEXT	 =  0,		/* alphanumeric modes		*/
		GT_1BIT	 =  1,		/*  1 bpp graphics		*/
		GT_4BIT	 =  2,		/*  4 bpp graphics		*/
		GT_8BIT	 =  3,		/*  8 bpp graphics		*/
		GT_15BIT =  4,		/* 15 bpp graphics 		*/
		GT_16BIT =  5,		/* 16 bpp graphics		*/
		GT_24BIT =  6,		/* 24 bpp graphics		*/
		GT_32BIT =  7,		/* 24 bpp word aligned  	*/
		GT_4BIT_PLANES =  8,	/*  4 bpp VGA plane graphics	*/
		GT_LAST  =  9		/* last mode assigned   	*/
	} ggi_graphtype;

	#define MASK(X) GM_##X = (1 << GT_##X) 	
	typedef enum
	{	MASK(TEXT),
		MASK(1BIT),
		MASK(4BIT),
		MASK(8BIT),
		MASK(15BIT),
		MASK(16BIT),
		MASK(24BIT),
		MASK(32BIT),
		MASK(4BIT_PLANES),
		GM_ALL	=  ((1 << GT_LAST) - 1)
	} ggi_graphmask;
	#undef MASK

	#ifdef NEED_GRAPHTYPE_SYMS
          static char *ggi_graphmask_sym[GT_LAST] =
	  {"GM_TEXT",  "GM_1BIT", "GM_4BIT", "GM_8BIT", "GM_15BIT", "GM_16BIT",
           "GM_24BIT", "GM_32BIT", "GM_4BIT_PLANES" };

	  #define GRAPHMASK(X) (ggi_graphmask_sym[ X ])
	#endif

	typedef enum
	{
		ufClear = 0,	ufAnd,		ufAndReverse,	ufCopy,
		ufAndInverted,	ufNoOp,		ufXor,		ufOr,
		ufNor,		ufEquiv,	ufInvert,	ufOrReverse,
		ufCopyInverted,	ufOrInverted,	ufNand,		ufSet
	} ggi_update_fn;

	typedef struct		/* requested by user and changed by driver */
	{
	  ggi_coord		visible;	/* vis. pixels, not changed   */
          ggi_coord		virtual;	/* virt. pixels, may change   */
	  ggi_coord		size;		/* size of visible in mm      */
	  ggi_graphtype		graphtype;	/* which mode ?		      */
#if 0
	  unsigned short	pages;		/* min. # of graphics pages   */
	  unsigned short	page0y;		/* size of graphics page 0    */
#endif
	  ggi_coord		textgrid;	/* fontsize 		      */
	} ggi_mode;

	typedef struct
	{
	  unsigned char		*data;		/* font bitmap data	      */
	  unsigned char		xsize, ysize;	/* size of one character cell */
	  unsigned char		cr_start,	/* cursor start line in cell  */
				cr_end;		/* cursor end line in cell    */
	  unsigned short 	cursor;		/* cursor char & attribute    */

	  unsigned char		underline;	/* underline location in cell */
	  unsigned char		first;		/* first char of the font     */
	  unsigned char		last;		/* number of chars in font    */
	  unsigned char		free_data;	/* 1 if data was allocated    */
	} ggi_font;

	typedef struct
	{
	  int			first;		/* first index to global CLUT */
	  int			cnt;		/* (max) number of entries    */
	  ggi_color		*data;		/* rgb triples		      */
	} ggi_palette;

	typedef struct
	{ int self,shmid,len;
          void *shmat;
	} ramdaemon_shm;
#endif	/* #ifdef _ggi_h */
