/* 
   Graphics library for GGI. Horizontal lines.

   Copyright (C) 1995 Andreas Beck - becka@hp.rz.uni-duesseldorf.de

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _LIB_COMPILE_

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <signal.h>
#include <sys/fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <string.h>
#include <ggi/graphlib.h>


/**********************************/
/* draw/get/put a horizontal line */
/**********************************/

static short bytemask[8]={0xff,0x7f,0x3f,0x1f,0xf,0x7,0x3,0x1}; 

static int drawhline4bit_pl(int unused,void *xyw)
{ 
  char *adr;int i;
  volatile register char foo;
  struct { int x,y,w; } *par;
  
  set16col(ggi_drwinfo.fg_color);
  par=xyw;
  adr=(char *)(ggi_membuffer+(((par->y*CurrentMode.virtual.x)+par->x)>>3));
  
  if ((i=par->x&7)) { foo=*adr; *adr++=bytemask[i]; } 

  for ( i=par->w-i; i>=8; i-=8) *adr++=0xff;
    
  if (i) { foo=*adr; *adr=~bytemask[i]; };
  
  return 0;
}

static int drawhline1bit(int unused,void *xyw)
{ 
  char *adr;int i;
  struct { int x,y,w; } *par;
  
  par=xyw;
  adr=(char *)(ggi_membuffer+(((par->y*CurrentMode.virtual.x)+par->x)>>3));

  if (ggi_drwinfo.fg_color&1)
    {
      if ((i=par->x&7)) *adr++|=bytemask[i];
      for (i=par->w-i;i>=8;i-=8) *adr++=0xff;
      if (i) *adr|=~bytemask[i];
    } else {
      if ((i=par->x&7)) *adr++&=~bytemask[i];
      for (i=par->w-i;i>=8;i-=8) *adr++=0x00;
      if (i) *adr&=bytemask[i];
    };
  return 0;
}


static int drawhline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;

	par=xyw;
	memset(ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
	       ggi_drwinfo.fg_color,par->w);
  return 0;
}

static int drawhline4bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;
	int col;

	par=xyw;
	if (par->x&1) { ggi_drawpixel(par->x,par->y); par->x++; par->w-- ;}
	col=ggi_drwinfo.fg_color&0x0f;col+=col<<4;
	memset(ggi_membuffer+((par->y*CurrentMode.virtual.x+par->x)>>1),
	       col,(par->w>>1));
	if (par->w&1) ggi_drawpixel(par->x+par->w-1,par->y);
	       
  return 0;
}

static int drawhline16bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;
	short *bla;int color,i;
	par=xyw;

	bla=(short *)ggi_membuffer+par->y*CurrentMode.virtual.x+par->x;
	if(par->x&1) { *bla++=ggi_drwinfo.fg_color; par->w--; };
	color=((ggi_drwinfo.fg_color&0xffff)<<16)|
	  (ggi_drwinfo.fg_color&0xffff);
	for (i=par->w;i>=2;i-=2) *((int *)bla)++=color;
	if (i) *bla=ggi_drwinfo.fg_color;
	return 0;
}

static int drawhline24bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;
	int *bla,*colors;char *foo;
	par=xyw;
	foo=(char *)ggi_membuffer+(par->y*CurrentMode.virtual.x+par->x)*3;
	while(((par->x)&3)&&par->w)
	  {  
	    *foo++=(ggi_drwinfo.fg_color)&0xFF;
	    *foo++=(ggi_drwinfo.fg_color>>8)&0xFF;
	    *foo++=(ggi_drwinfo.fg_color>>16)&0xFF;
	    par->x++;par->w--;
	  };
	colors=(int *)foo;
	if (par->w)
	  do {
	    *foo++=(ggi_drwinfo.fg_color)&0xFF;
	    *foo++=(ggi_drwinfo.fg_color>>8)&0xFF;
	    *foo++=(ggi_drwinfo.fg_color>>16)&0xFF;
	    par->x++;par->w--;
	  } while (((par->x)&3)&&par->w);
	bla=(int *)foo;
	while (par->w>3)
	  {
	    *bla++=colors[0];
	    *bla++=colors[1];
	    *bla++=colors[2];
	    par->w-=4;
	  };
	foo=(char *)bla;
	while (par->w)
	  {
            *foo++=(ggi_drwinfo.fg_color)&0xFF;
            *foo++=(ggi_drwinfo.fg_color>>8)&0xFF;
            *foo++=(ggi_drwinfo.fg_color>>16)&0xFF;
            par->w--;
	  };
  return 0;
}

static int drawhline32bit(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;
	int *bla;
	par=xyw;

	bla=(int *)ggi_membuffer+par->y*CurrentMode.virtual.x+par->x;
	while(par->w--) *bla++=ggi_drwinfo.fg_color;
	return 0;
}

static int drawhline_lower(int unused,void *xyw)
{ 
	struct { int x,y,w; } *par;

	par=xyw;
	while(par->w--) 
		ggi_drawpixel(par->x++,par->y);
	return(0);
}

int ggi_drawhline(int x,int y,int width)
{ 
	switchavail(GRAPHICS_DRAWHLINE,&x);

	switcha(GRAPHICS_DRAWHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
  		{ case GT_8BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline8bit,&x);
  		  case GT_16BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline16bit,&x);
                  case GT_24BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline24bit,&x);
  		  case GT_32BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline32bit,&x);
  		  case GT_4BIT_PLANES	:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline4bit_pl,&x);
  		  case GT_4BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline4bit,&x);
  		  case GT_1BIT		:setdoemu(GRAPHICS_DRAWHLINE,AVE_MODE,drawhline1bit,&x);
  		  default		:break;/* Use lower - no others now */
		}
	  case AVM_LOWER:
	  	setdoemu(GRAPHICS_DRAWHLINE,AVE_ALWAYS,drawhline_lower,&x);
   		break;
	  default       :
	  	ggi_panic("hline - no method.");
	  	break;
  }
  return(0);
}

static int puthline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w;
		 char *buffer; } *par;
	char *mem;

	par=xyw;
#if 0
	memcpy(ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
		par->buffer,par->w);
#else
	mem=ggi_membuffer+par->y*CurrentMode.virtual.x+par->x;
	while(par->w--)
	{ *mem++=*par->buffer++; }
#endif
	
	return 0;
}

static int puthline_lower(int unused,void *xyw)
{	
	struct { int x,y,w;
		 char *buffer; } *par;
	
	par=xyw;
  	while(par->w--) 
  		ggi_putpixel(par->x++,par->y,par->buffer++);
	return 0;
}

int ggi_puthline(int x,int y,int width,void *buffer)
{ 
	switchavail(GRAPHICS_PUTHLINE,&x);

	switcha(GRAPHICS_PUTHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
  		{ case GT_8BIT:setdoemu(GRAPHICS_PUTHLINE,AVE_MODE,puthline8bit,&x);
	  	  default:break;
	  	}
	  case AVM_LOWER:setdoemu(GRAPHICS_PUTHLINE,AVE_ALWAYS,puthline_lower,&x);
	  default       :
	  	ggi_panic("hline - no method.");break;
	}
	return(0);
}

static int gethline8bit(int unused,void *xyw)
{ 
	struct { int x,y,w;
		 char *buffer; } *par;

	par=xyw;
	memcpy(par->buffer,ggi_membuffer+par->y*CurrentMode.virtual.x+par->x,
		par->w);

	return 0;
}

static int gethline_lower(int unused,void *xyw)
{	
	struct { int x,y,w;
		 char *buffer; } *par;
	
	par=xyw;
  	while(par->w--) 
  		ggi_getpixel(par->x++,par->y,par->buffer++);
	return 0;
}

int ggi_gethline(int x,int y,int width,void *buffer)
{
	switchavail(GRAPHICS_GETHLINE,&x);

	switcha(GRAPHICS_GETHLINE)
	{ case AVM_MMAP :
		switch(CurrentMode.graphtype)
		{ case GT_8BIT:setdoemu(GRAPHICS_GETHLINE,AVE_MODE,gethline8bit,&x);
  		  default: break;
  		}
	  case AVM_LOWER:setdoemu(GRAPHICS_GETHLINE,AVE_ALWAYS,gethline_lower,&x);
	  default       :
	  	ggi_panic("hline - no method.");break;
	}
	return(0);
}

